//
// Created by Emil Bjoerlykhaug 06.04.16
//

#include "hahaha/computing.hpp"


/**
 * @brief Constructor of the Computing class. Takes no arguments
 */
Computing::Computing(){

}

/**
 * @brief Calculate VFH descriptors for each point cloud in the input cluster
 * @param clusters The input cluster to calculate the descriptors for
 * @return A vector with the descriptors for the point cloud in the cluster
 */
std::vector<pcl::PointCloud<pcl::VFHSignature308>::Ptr> Computing::calculateVFHDescriptors(std::vector<pcl::PointCloud<pcl::PointXYZ>::Ptr> clusters)
{
    //Debug
    std::cout << "Hi!" << std::endl;
    //Create a variable for the descriptors
    std::vector<pcl::PointCloud<pcl::VFHSignature308>::Ptr> descriptors;

    //Iterate through every item in the cluster, and calculate descriptors for it
    for(int i=0;i<clusters.size();i++){
        // Object for storing the normals.
        pcl::PointCloud<pcl::Normal>::Ptr normals(new pcl::PointCloud<pcl::Normal>);
        // Object for storing the VFH descriptor.
        pcl::PointCloud<pcl::VFHSignature308>::Ptr descriptor(new pcl::PointCloud<pcl::VFHSignature308>);

        // Estimate the normals.
        pcl::NormalEstimation<pcl::PointXYZ, pcl::Normal> normalEstimation;
        normalEstimation.setInputCloud(clusters[i]);
        normalEstimation.setRadiusSearch(0.03);
        pcl::search::KdTree<pcl::PointXYZ>::Ptr kdtree(new pcl::search::KdTree<pcl::PointXYZ>);
        normalEstimation.setSearchMethod(kdtree);
        normalEstimation.compute(*normals);

        // VFH estimation object.
        pcl::VFHEstimation<pcl::PointXYZ, pcl::Normal, pcl::VFHSignature308> vfh;
        vfh.setInputCloud(clusters[i]);
        vfh.setInputNormals(normals);
        vfh.setSearchMethod(kdtree);
        // Optionally, we can normalize the bins of the resulting histogram,
        // using the total number of points.
        vfh.setNormalizeBins(true);
        // Also, we can normalize the SDC with the maximum size found between
        // the centroid and any of the cluster's points.
        vfh.setNormalizeDistance(false);
        //Compute the descriptor
        vfh.compute(*descriptor);
        descriptors.push_back(descriptor);
    }
    return descriptors;
}

/**
 * @brief Passthrough filter on the input cloud
 * @param cloud_before_filter The input cloud
 * @param zFilter Bool if the filter also should filter in the z-axis
 * @return the filtered cloud
 */
pcl::PointCloud<pcl::PointXYZ>::Ptr Computing::passThroughFilter(pcl::PointCloud<pcl::PointXYZ>::Ptr cloud_before_filter, bool zFilter)
{
    //Initialize variables
    pcl::PointCloud<pcl::PointXYZ>::Ptr cloud(new pcl::PointCloud<pcl::PointXYZ>);
    pcl::PointCloud<pcl::PointXYZ>::Ptr cloud1(new pcl::PointCloud<pcl::PointXYZ>);
    pcl::PointCloud<pcl::PointXYZ>::Ptr cloud2(new pcl::PointCloud<pcl::PointXYZ>);
    // Pass through filter
    pcl::PassThrough<pcl::PointXYZ> pass;
    pass.setInputCloud (cloud_before_filter);
    //pass.setFilterFieldName ("z");
    //pass.setFilterLimits (0, 2.7);
    //Filter in the y-direction
    pass.setFilterFieldName ("y");
    pass.setFilterLimits (-2, 0);

    pass.filter (*cloud1);
    //Filter in the x-direction
    pass.setInputCloud(cloud1);
    pass.setFilterFieldName ("x");
    pass.setFilterLimits (-0.45, 0.45);
    pass.filter (*cloud2);

    //Check if the z-direction should be filtered as well.
    if(zFilter){
        pass.setInputCloud(cloud2);
        pass.setFilterFieldName ("z");
        pass.setFilterLimits (0, 2.350);
        pass.filter (*cloud);
    }
    else{
        cloud = cloud2;
    }

    return cloud;
}

/**
 * @brief Downsample an input cloud using a voxel grid
 * @param cloud The input cloud
 * @param leaf_size The size of the voxels. All the points inside a voxel will be replaced with the centeroid
 * @return The downsampled cloud
 */
pcl::PointCloud<pcl::PointXYZ>::Ptr Computing::downsample_voxel(pcl::PointCloud<pcl::PointXYZ>::Ptr cloud, float leaf_size)
{
    pcl::VoxelGrid<pcl::PointXYZ> vg;
    pcl::PointCloud<pcl::PointXYZ>::Ptr cloud_filtered (new pcl::PointCloud<pcl::PointXYZ>);
    vg.setInputCloud (cloud);
    vg.setLeafSize (leaf_size, leaf_size, leaf_size);
    vg.filter (*cloud_filtered);
    return cloud_filtered;
}

/**
 * @brief A method which will remove the biggest planar face in a point cloud.
 * @param inCloud The input cloud
 * @return The point cloud without the planar face
 */
pcl::PointCloud<pcl::PointXYZ>::Ptr  Computing::segmentation(pcl::PointCloud<pcl::PointXYZ>::Ptr inCloud)
{

    pcl::PointCloud<pcl::PointXYZ>::Ptr cloud_f(new pcl::PointCloud<pcl::PointXYZ>);
    // Create the segmentation object for the planar model and set all the parameters
    pcl::SACSegmentation<pcl::PointXYZ> seg;
    pcl::PointIndices::Ptr inliers (new pcl::PointIndices);
    pcl::ModelCoefficients::Ptr coefficients (new pcl::ModelCoefficients);
    pcl::PointCloud<pcl::PointXYZ>::Ptr cloud_plane (new pcl::PointCloud<pcl::PointXYZ> ());
    seg.setOptimizeCoefficients (true);
    seg.setModelType (pcl::SACMODEL_PLANE);
    seg.setMethodType (pcl::SAC_RANSAC);
    seg.setMaxIterations (100);
    seg.setDistanceThreshold (0.01);


    seg.setInputCloud (inCloud);
    seg.segment (*inliers, *coefficients);
    //Check if there are not planar faces in the model
    if (inliers->indices.size () == 0)
    {
        std::cout << "Could not estimate a planar model for the given dataset." << std::endl;
        //break;
    }

    // Extract the planar inliers from the input cloud
    pcl::ExtractIndices<pcl::PointXYZ> extract;
    extract.setInputCloud (inCloud);
    extract.setIndices (inliers);
    extract.setNegative (false);

    // Get the points associated with the planar surface
    extract.filter (*cloud_plane);
    std::cout << "PointCloud representing the planar component: " << cloud_plane->points.size () << " data points." << std::endl;

    // Remove the planar inliers, extract the rest
    extract.setNegative (true);
    extract.filter (*cloud_f);
    return cloud_f;

}

/**
 * @brief Divide an input point cloud into clusters of points
 * @param inCloud The input point cloud
 * @return the clusters
 */
std::vector<pcl::PointCloud<pcl::PointXYZ>::Ptr>  Computing::extract_clusters(pcl::PointCloud<pcl::PointXYZ>::Ptr inCloud)
{
    std::vector<pcl::PointCloud<pcl::PointXYZ>::Ptr> clusters;
    // Creating the KdTree object for the search method of the extraction
    pcl::search::KdTree<pcl::PointXYZ>::Ptr tree (new pcl::search::KdTree<pcl::PointXYZ>);
    tree->setInputCloud (inCloud);

    std::vector<pcl::PointIndices> cluster_indices;
    pcl::EuclideanClusterExtraction<pcl::PointXYZ> ec;
    ec.setClusterTolerance (0.01); // 2cm
    ec.setMinClusterSize (200);
    ec.setMaxClusterSize (25000);
    ec.setSearchMethod (tree);
    ec.setInputCloud (inCloud);
    ec.extract (cluster_indices);

    //Put all the point cloud in a vector, the cluster vector. Also prints the size of the each cluster
    for(int i = 0; i< cluster_indices.size(); i++){
        pcl::PointCloud<pcl::PointXYZ>::Ptr tmpcloud (new pcl::PointCloud<pcl::PointXYZ>);
        pcl::copyPointCloud(*inCloud,cluster_indices[i],*tmpcloud);
        clusters.push_back(tmpcloud);
        std::cout << "PointCloud representing the Cluster: " << tmpcloud->size() << " data points." << std::endl;
    }
    return clusters;
}

/**
 * @brief A method for generating a search tree. A search tree is required for finding the correct cluster with the help of the VFH descriptor. Method is required for matching clouds
 * @param inDescriptor the descriptor to generate a search tree for
 * @return the search tree for given descriptor
 */
pcl::KdTreeFLANN<pcl::VFHSignature308>::Ptr Computing::generate_search_tree(std::vector<pcl::PointCloud<pcl::VFHSignature308>::Ptr> inDescriptor)
{
    // Take global descriptors and put them in a kdtree

    pcl::PointCloud<pcl::VFHSignature308>::Ptr global_descriptor (new pcl::PointCloud<pcl::VFHSignature308>());
    pcl::KdTreeFLANN<pcl::VFHSignature308>::Ptr search_tree(new pcl::KdTreeFLANN<pcl::VFHSignature308>);
        for(int i = 0; i< inDescriptor.size(); i++){

        pcl::PointCloud<pcl::VFHSignature308>::Ptr temp_descriptor = inDescriptor.at(i);
        *global_descriptor += *(temp_descriptor);
    }


    search_tree->setInputCloud(global_descriptor);
    std::cout << "Size of tree: " << global_descriptor->size() << std::endl;
    return (search_tree);
}

/**
 * @brief Match an input descriptor to a search tree made of a vector of descriptors. Will return which descriptor in the tree that best matches, and the square distance of that match
 * @param global_descriptor The descriptor to match
 * @param inTree A search tree of descriptors
 * @return A vector with descriptor in the search tree that best matches, and the square distance of that match
 */
std::vector<float> Computing::match_cloud(pcl::PointCloud<pcl::VFHSignature308>::Ptr global_descriptor, pcl::KdTreeFLANN<pcl::VFHSignature308>::Ptr inTree)
{
    std::vector<float> return_values;
    std::vector<int> best_match(1);
    std::vector<float> square_distance(1);

    inTree->nearestKSearch(global_descriptor->points[0],1,best_match,square_distance);
    return_values.push_back(best_match[0]);
    return_values.push_back(square_distance[0]);
    return return_values;
}

/**
 * @brief A method for finding the best match from two sets of descriptors.
 * @param cluster_descriptors Descriptors for one of the clusters
 * @param cad_descriptors Descriptors for the ray traced CAD model
 * @return Which of the models in the CAD cluster that best matches one of the descriptors in the cloud cluster
 */
int Computing::search_for_correct_cluster(std::vector<pcl::PointCloud<pcl::VFHSignature308>::Ptr> cluster_descriptors, std::vector<pcl::PointCloud<pcl::VFHSignature308>::Ptr> cad_descriptors)
{
    //Generate a search tree from the CAD descriptors
    pcl::KdTreeFLANN<pcl::VFHSignature308>::Ptr tree = generate_search_tree(cad_descriptors);

    float min_distance = 10000;
    int correct_cluster = 50;
    std::vector<float> search_result;

    //For each of the descriptors in the cloud cluster, match it against the search tree from the CAD descriptors
    for (int i = 0; i < cluster_descriptors.size(); i++){
        search_result = match_cloud(cluster_descriptors.at(i),tree);
        std::cout << search_result.at(1) << std::endl;
        //If the new search result is less than the previous minimum distance, change the new minimum distance to this one. And save what is the new correct cluster
        if(search_result[1] < min_distance){
            min_distance = search_result[1];
            correct_cluster = i;
        }
    }
    std::cout << "Correct cluster: " << correct_cluster << std::endl;
    return correct_cluster;

}

/**
 * @brief Compute FPFH descriptor for a input cloud. Normals to the cloud, keypoints, and search radius is necessary as inputs
 * @param cloud The input cloud
 * @param normals The normals to the cloud
 * @param keypoint Keypoints in the cloud
 * @param radius The search radius for calculating the descriptor
 * @return The caculated descriptors
 */
pcl::PointCloud<pcl::FPFHSignature33>::Ptr Computing::compute_FPFH(pcl::PointCloud<pcl::PointXYZ>::Ptr cloud,
                                                                   pcl::PointCloud<pcl::Normal>::Ptr normals, pcl::PointCloud<pcl::PointXYZ>::Ptr keypoint, float radius)
{
    //Initiate object
    pcl::FPFHEstimationOMP<pcl::PointXYZ,pcl::Normal,pcl::FPFHSignature33> fpfh_obj;
    //To speed up the process
    fpfh_obj.setNumberOfThreads(8);
    fpfh_obj.setSearchMethod(pcl::search::Search<pcl::PointXYZ>::Ptr (new pcl::search::KdTree<pcl::PointXYZ>));
    fpfh_obj.setRadiusSearch(radius);
    fpfh_obj.setSearchSurface(cloud);
    fpfh_obj.setInputNormals(normals);
    fpfh_obj.setInputCloud(keypoint);
    pcl::PointCloud<pcl::FPFHSignature33>::Ptr descriptors(new pcl::PointCloud<pcl::FPFHSignature33>());
    fpfh_obj.compute(*descriptors);
    return descriptors;

}

/**
 * @brief A method to find keypoints in a input point cloud
 * @param inCloud The input cloud
 * @param min_scale Minimum scale
 * @param nr_octaves Number of octaves
 * @param nr_scale_pr_octave Number of scale per octave
 * @return
 */
pcl::PointCloud<pcl::PointXYZ>::Ptr Computing::generate_keypoints_sift(pcl::PointCloud<pcl::PointXYZ>::Ptr inCloud, float min_scale, int nr_octaves, int nr_scale_pr_octave)
{
    pcl::PointCloud<pcl::PointXYZRGB>::Ptr rgbcloud(new pcl::PointCloud<pcl::PointXYZRGB>);
    pcl::copyPointCloud(*inCloud,*rgbcloud);
    for(int i = 0; i< rgbcloud->size(); i++){
        rgbcloud->points[i].r = 255;
        rgbcloud->points[i].g = 255;
        rgbcloud->points[i].b = 255;
    }

    pcl::SIFTKeypoint<pcl::PointXYZRGB, pcl::PointWithScale> sift;
    sift.setSearchMethod(pcl::search::Search<pcl::PointXYZRGB>::Ptr (new pcl::search::KdTree<pcl::PointXYZRGB>));
    sift.setScales(min_scale,nr_octaves,nr_scale_pr_octave);
    sift.setInputCloud(rgbcloud);
    sift.setMinimumContrast(0.0);
    pcl::PointCloud<pcl::PointWithScale> tmp;
    sift.compute(tmp);
    pcl::PointCloud<pcl::PointXYZ>::Ptr keyPoints (new pcl::PointCloud<pcl::PointXYZ>);
    pcl::copyPointCloud(tmp, *keyPoints);
    return keyPoints;
}

/**
 * @brief Aligns two point clouds with feature matching
 * @param source_cloud The source cloud
 * @param target_cloud The target cloud
 * @param min_sample Minimum sample
 * @param max_corr_dist Maximum distance between correspondences
 * @return The transformation matrix
 */
Eigen::Matrix4f Computing::alignment(pcl::PointCloud<pcl::PointXYZ>::Ptr source_cloud, pcl::PointCloud<pcl::PointXYZ>::Ptr target_cloud, int min_sample, int max_corr_dist)
{
    // Sample consensus alignment
    pcl::SampleConsensusInitialAlignment<pcl::PointXYZ,pcl::PointXYZ,pcl::FPFHSignature33> initial_alignment;

    initial_alignment.setMinSampleDistance(min_sample);
    initial_alignment.setMaxCorrespondenceDistance(max_corr_dist);
    initial_alignment.setMaximumIterations(100);


    //Compute keypoints
    pcl::PointCloud<pcl::PointXYZ>::Ptr source_key = generate_keypoints_sift(source_cloud,0.001,3,3);
    std::cout << source_cloud->size() << std::endl;
    std::cout << source_key->size() << std::endl;
    // Estimate the normals.
    pcl::PointCloud<pcl::Normal>::Ptr normals1(new pcl::PointCloud<pcl::Normal>);
    pcl::NormalEstimation<pcl::PointXYZ, pcl::Normal> normalEstimation1;
    normalEstimation1.setInputCloud(source_cloud);
    normalEstimation1.setRadiusSearch(0.03);
    pcl::search::KdTree<pcl::PointXYZ>::Ptr kdtree1(new pcl::search::KdTree<pcl::PointXYZ>);
    normalEstimation1.setSearchMethod(kdtree1);
    normalEstimation1.compute(*normals1);
    //Compute local descriptors
    pcl::PointCloud<pcl::FPFHSignature33>::Ptr source_descriptors = compute_FPFH(source_cloud, normals1, source_key,0.15);
    initial_alignment.setInputSource(source_key);
    initial_alignment.setSourceFeatures(source_descriptors);


    //Compute keypoints
    pcl::PointCloud<pcl::PointXYZ>::Ptr target_key = generate_keypoints_sift(target_cloud,0.001,3,3);
    std::cout << target_cloud->size() << std::endl;
    std::cout << target_key->size() << std::endl;
    // Estimate the normals.
    pcl::PointCloud<pcl::Normal>::Ptr normals2(new pcl::PointCloud<pcl::Normal>);
    pcl::NormalEstimation<pcl::PointXYZ, pcl::Normal> normalEstimation2;
    normalEstimation2.setInputCloud(target_cloud);
    normalEstimation2.setRadiusSearch(0.03);
    pcl::search::KdTree<pcl::PointXYZ>::Ptr kdtree2(new pcl::search::KdTree<pcl::PointXYZ>);
    normalEstimation2.setSearchMethod(kdtree2);
    normalEstimation2.compute(*normals2);
    //Compute local descriptors
    pcl::PointCloud<pcl::FPFHSignature33>::Ptr target_descriptors = compute_FPFH(target_cloud, normals2, target_key,0.15);

    initial_alignment.setInputTarget(target_key);
    initial_alignment.setTargetFeatures(target_descriptors);


    pcl::PointCloud<pcl::PointXYZ> registration_output;
    initial_alignment.align(registration_output);

    std::cout << initial_alignment.getFinalTransformation() << std::endl;

    return (initial_alignment.getFinalTransformation());
}

/**
 * @brief Aligns two point cloud with ICP
 * @param source_cloud Source cloud
 * @param target_cloud Target cloud
 * @return The transformation matrix
 */
Eigen::Matrix4f Computing::icp_alignment(pcl::PointCloud<pcl::PointXYZ>::Ptr source_cloud, pcl::PointCloud<pcl::PointXYZ>::Ptr target_cloud)
{
    // Compute the ICP alignment
    pcl::PointCloud<pcl::PointXYZ>::Ptr finalCloud(new pcl::PointCloud<pcl::PointXYZ>);

    pcl::IterativeClosestPoint<pcl::PointXYZ, pcl::PointXYZ> icp_b;
    icp_b.setInputSource(source_cloud);
    icp_b.setInputTarget(target_cloud);
    icp_b.setMaximumIterations(100);

    icp_b.align(*finalCloud);
    if (icp_b.hasConverged())
    {
        std::cout << "ICP converged." << std::endl
                  << "The score is " << icp_b.getFitnessScore() << std::endl;
        std::cout << "Transformation matrix:" << std::endl;
        std::cout << icp_b.getFinalTransformation() << std::endl;
    }
    else std::cout << "ICP did not converge." << std::endl;
    return icp_b.getFinalTransformation();
}

/**
 * @brief Finds the differences between two point cloud using the SegmentDifferences method in PCL
 * @param source_cloud Source point cloud
 * @param target_cloud Target point cloud
 * @param distThresh The distance threshold
 * @return the resulting point cloud
 */
pcl::PointCloud<pcl::PointXYZ>::Ptr Computing::find_differences(pcl::PointCloud<pcl::PointXYZ>::Ptr source_cloud, pcl::PointCloud<pcl::PointXYZ>::Ptr target_cloud, float distThresh)
{
    // The output cloud
    pcl::PointCloud<pcl::PointXYZ>::Ptr segmentedCloud(new pcl::PointCloud<pcl::PointXYZ>);

    pcl::search::KdTree<pcl::PointXYZ>::Ptr tree(new pcl::search::KdTree<pcl::PointXYZ>);
    pcl::SegmentDifferences<pcl::PointXYZ> seg_diff;
    seg_diff.setSearchMethod(tree);
    seg_diff.setDistanceThreshold(distThresh);
    seg_diff.setInputCloud(source_cloud);
    seg_diff.setTargetCloud(target_cloud);
    seg_diff.segment(*segmentedCloud);
    return segmentedCloud;
}

/**
 * @brief A statistical outlier removal filter
 * @param inCloud The cloud to process
 * @param meanK Global distance mean deviation
 * @param stdDev Standard deviation of distance
 * @return Point cloud with noise filtered
 */
pcl::PointCloud<pcl::PointXYZ>::Ptr Computing::statOutlierRemoval(pcl::PointCloud<pcl::PointXYZ>::Ptr inCloud, int meanK, float stdDev)
{
    pcl::PointCloud<pcl::PointXYZ>::Ptr cloud_filtered (new pcl::PointCloud<pcl::PointXYZ>);

    std::cout << "Cloud before filtering: " << std::endl;
    std::cout << *inCloud << std::endl;

    // Create the filtering object
    pcl::StatisticalOutlierRemoval<pcl::PointXYZ> sor;
    sor.setInputCloud (inCloud);
    sor.setMeanK (meanK);
    sor.setStddevMulThresh (stdDev);
    sor.filter (*cloud_filtered);

    //Print to see the difference in points after filtering
    std::cout << "Cloud after filtering: " << std::endl;
    std::cout << *cloud_filtered << std::endl;

    return cloud_filtered;
}

/**
 * @brief Calculates normals for a input point cloud
 * @param inCloud The input cloud
 * @return The normals for the input cloud
 */
pcl::PointCloud<pcl::Normal>::Ptr Computing::calculateNormals(pcl::PointCloud<pcl::PointXYZ>::Ptr inCloud)
{
    pcl::PointCloud<pcl::Normal>::Ptr normals1(new pcl::PointCloud<pcl::Normal>);
    pcl::NormalEstimation<pcl::PointXYZ, pcl::Normal> normalEstimation1;
    normalEstimation1.setInputCloud(inCloud);
    normalEstimation1.setRadiusSearch(0.2);
    pcl::search::KdTree<pcl::PointXYZ>::Ptr kdtree1(new pcl::search::KdTree<pcl::PointXYZ>);
    normalEstimation1.setSearchMethod(kdtree1);
    normalEstimation1.compute(*normals1);


    return normals1;
}

/**
 * @brief Calculate normals for a point cloud, given a viewpoint
 * @param inCloud The input cloud
 * @param x X-coordinate for the viewpoint
 * @param y Y-coordinate for the viewpoint
 * @param z Z-coordinate for the viewpoint
 * @return The calculated normals
 */
pcl::PointCloud<pcl::Normal>::Ptr Computing::calculateNormalsGivenViewPoint(pcl::PointCloud<pcl::PointXYZ>::Ptr inCloud,float x, float y, float z)
{
    pcl::PointCloud<pcl::Normal>::Ptr normals1(new pcl::PointCloud<pcl::Normal>);
    pcl::NormalEstimation<pcl::PointXYZ, pcl::Normal> normalEstimation1;
    normalEstimation1.setInputCloud(inCloud);
    normalEstimation1.setRadiusSearch(0.03);
    normalEstimation1.setViewPoint(x,y,z);
    pcl::search::KdTree<pcl::PointXYZ>::Ptr kdtree1(new pcl::search::KdTree<pcl::PointXYZ>);
    normalEstimation1.setSearchMethod(kdtree1);
    normalEstimation1.compute(*normals1);


    return normals1;
}

/**
 * @brief Flips the direction of the normals in the input
 * @param inNormal The normals to flip
 * @return The flipped normals
 */
pcl::PointCloud<pcl::Normal>::Ptr Computing::flipNormals(pcl::PointCloud<pcl::Normal>::Ptr inNormal)
{
    std::cout << "test" << std::endl;
    pcl::PointCloud<pcl::Normal>::Ptr outNormal = inNormal;
    for (int i=0;i<inNormal->size();i++){
        //std::cout << "x: "<< outNormal->points[i].normal_x << std::endl;
        //std::cout << "y: "<< outNormal->points[i].normal_y << std::endl;
        //std::cout << "z: "<< outNormal->points[i].normal_z << std::endl;
        outNormal->points[i].normal_x = -1.0*outNormal->points[i].normal_x;
        outNormal->points[i].normal_y = -1.0*outNormal->points[i].normal_y;
        outNormal->points[i].normal_z = -1.0*outNormal->points[i].normal_z;
        //std::cout << "x2: "<< outNormal->points[i].normal_x << std::endl;
        //std::cout << "y2: "<< outNormal->points[i].normal_y << std::endl;
        //std::cout << "z2: "<< outNormal->points[i].normal_z << std::endl;
    }
    return outNormal;
}

/**
 * @brief Calculates new points at a given distance from the input point, along the input normal
 * @param inCloud The input cloud
 * @param inNormal The normals to the input cloud
 * @param distance The distance from the old point the new point should be
 * @return The new points, organized in a cloud
 */
pcl::PointCloud<pcl::PointXYZ>::Ptr Computing::calculateNewPointsFromPCandNormals(pcl::PointCloud<pcl::PointXYZ>::Ptr inCloud, pcl::PointCloud<pcl::Normal>::Ptr inNormal, float distance)

{
    pcl::PointCloud<pcl::PointXYZ>::Ptr outCloud (new pcl::PointCloud<pcl::PointXYZ>);
    pcl::PointCloud<pcl::Normal>::Ptr outNormal (new pcl::PointCloud<pcl::Normal>);
    *outCloud = *inCloud;
    *outNormal = *inNormal;
    std::cout << "test" << std::endl;
    for (int i=0; i < inCloud->size();i++){
        std::cout << "x: "<< outCloud->points[i].x << std::endl;
        outCloud->points[i].x = outCloud->points[i].x+outNormal->points[i].normal_x*distance;
        std::cout << "x: "<< outNormal->points[i].normal_x << std::endl;
        std::cout << distance << std::endl;
        std::cout << "x: "<< outCloud->points[i].x << std::endl;
        outCloud->points[i].y = outCloud->points[i].y+outNormal->points[i].normal_y*distance;
        outCloud->points[i].z = outCloud->points[i].z+outNormal->points[i].normal_z*distance;
    }
    return outCloud;

}


